@@ -1,16 +1,23 @@
-#  dmt(Z, mean = rep(0,133), S = (nu - 2)/nu*(model$sigma2 * cov_gen(X, theta = model$theta) + diag(model$g, 133)), log = T)
-loglik_HomTP <- function(X0, Z0, Z, mult, theta, g, nu, sigma2, beta0 = 0, covtype = "Gaussian", eps = sqrt(.Machine$double.eps)){
+#  dmt(Z, mean = rep(0,133), S = (nu - 2)/nu*(model$sigma2 * cov_gen(X, theta = model$theta) + diag(model$g, 133)), df = model$nu, log = T)
+loglik_HomTP <- function(X0, Z0, Z, mult, theta, g, nu, sigma2, beta0 = 0,
+                         covtype = "Gaussian", eps = sqrt(.Machine$double.eps), env = NULL){
   n <- nrow(X0)
   N <- length(Z)
   
+  C <- cov_gen(X1 = X0, theta = theta, type = covtype)
   # Temporarily store Cholesky transform of K in Ki
-  Ki <- chol(add_diag(sigma2 * cov_gen(X1 = X0, theta = theta, type = covtype), eps + g/mult))
+  Ki <- chol(add_diag(sigma2 * C, eps + g/mult))
   ldetKi <- - 2 * sum(log(diag(Ki))) # log determinant from Cholesky
   Ki <- chol2inv(Ki)
   
+  if(!is.null(env)){
+    env$C <- C
+    env$Ki <- Ki
+  } 
+  
   psi_0 <- drop(crossprod(Z0 - beta0, Ki) %*% (Z0 - beta0))
   
-  psi <- (crossprod(Z - beta0, Z - beta0) - crossprod((Z0 - beta0) * mult, Z0 - beta0))/g + psi_0
+  psi <- (crossprod(Z - beta0) - crossprod((Z0 - beta0) * mult, Z0 - beta0))/g + psi_0
   
   return(-N/2 * log((nu - 2) * pi) + ldetKi/2 - (N - n)/2 * log(g) - 1/2 * sum(log(mult)) + lgamma((nu + N)/2) - lgamma(nu/2) - (nu + N)/2 * log(1 + psi/(nu - 2)))
 }
@@ -29,12 +36,18 @@ loglik_HomTP <- function(X0, Z0, Z, mult, theta, g, nu, sigma2, beta0 = 0, covty
 # nu degrees of freedom
 # beta0 trend
 ## @return gradient with respect to theta and g
-dlogLikHomTP <- function(X0, Z0, Z, mult, theta, g, nu, sigma2, beta0 = 0, covtype = "Gaussian", eps = sqrt(.Machine$double.eps), components = c("theta", "g", "nu", "sigma2")){
+dlogLikHomTP <- function(X0, Z0, Z, mult, theta, g, nu, sigma2, beta0 = 0, covtype = "Gaussian", eps = sqrt(.Machine$double.eps),
+                         components = c("theta", "g", "nu", "sigma2"), env = NULL){
   N <- length(Z)
   n <- nrow(X0)
   
-  C <- cov_gen(X1 = X0, theta = theta, type = covtype)
-  Ki <- chol2inv(chol(sigma2*C + diag(eps + g / mult)))
+  if(!is.null(env)){
+    C <- env$C
+    Ki <- env$Ki
+  }else{
+    C <- cov_gen(X1 = X0, theta = theta, type = covtype)
+    Ki <- chol2inv(chol(sigma2*C + diag(eps + g / mult)))
+  }
   
   Z0 <- Z0 - beta0
   Z <- Z - beta0
@@ -43,7 +56,7 @@ dlogLikHomTP <- function(X0, Z0, Z, mult, theta, g, nu, sigma2, beta0 = 0, covty
   
   psi_0 <- drop(crossprod(Z0, KiZ0))
   
-  psi <- (crossprod(Z, Z) - crossprod(Z0 * mult, Z0))/g + psi_0
+  psi <- (crossprod(Z) - crossprod(Z0 * mult, Z0))/g + psi_0
   
   
   tmp1 <- tmp2 <- tmp3 <- tmp4 <- NULL
@@ -70,7 +83,7 @@ dlogLikHomTP <- function(X0, Z0, Z, mult, theta, g, nu, sigma2, beta0 = 0, covty
   
   # Fourth component derivative with respect to g
   if("g" %in% components)
-    tmp4 <- (nu + N)/(2 * (nu + psi - 2)) * ((crossprod(Z, Z) - crossprod(Z0 * mult, Z0))/g^2 + sum(KiZ0^2/mult)) - (N - n)/ (2*g) - 1/2 * sum(diag(Ki)/mult)
+    tmp4 <- (nu + N)/(2 * (nu + psi - 2)) * ((crossprod(Z) - crossprod(Z0 * mult, Z0))/g^2 + sum(KiZ0^2/mult)) - (N - n)/ (2*g) - 1/2 * sum(diag(Ki)/mult)
   
   return(c(tmp1, tmp2, tmp3, tmp4))
 }
@@ -220,7 +233,7 @@ mleHomTP <- function(X, Z, lower, upper, known = list(beta0 = 0),
   #   trendtype <- 'SK'
   
   ## General definition of fn and gr
-  fn <- function(par, X0, Z0, Z, mult, beta0, theta, sigma2, nu, g){
+  fn <- function(par, X0, Z0, Z, mult, beta0, theta, sigma2, nu, g, env){
     idx <- 1 # to store the first non used element of par
     
     if(is.null(theta)){
@@ -240,10 +253,10 @@ mleHomTP <- function(X, Z, lower, upper, known = list(beta0 = 0),
       g <- par[idx]
     }
     return(loglik_HomTP(X0 = X0, Z0 = Z0, Z = Z, mult = mult, theta = theta, g = g, nu = nu, sigma2 = sigma2,
-                        beta0 = beta0, covtype = covtype, eps = eps))
+                        beta0 = beta0, covtype = covtype, eps = eps, env = env))
   }
   
-  gr <- function(par, X0, Z0, Z, mult, beta0, theta, sigma2, nu, g){
+  gr <- function(par, X0, Z0, Z, mult, beta0, theta, sigma2, nu, g, env){
     idx <- 1
     components <- NULL
     
@@ -268,7 +281,7 @@ mleHomTP <- function(X, Z, lower, upper, known = list(beta0 = 0),
     }
     
     return(dlogLikHomTP(X0 = X0, Z0 = Z0, Z = Z, mult = mult, theta = theta, g = g, nu = nu, sigma2 = sigma2,
-                        beta0 = beta0, covtype = covtype, eps = eps, components = components))
+                        beta0 = beta0, covtype = covtype, eps = eps, components = components, env = env))
   }
   
   ## All known
@@ -303,8 +316,10 @@ mleHomTP <- function(X, Z, lower, upper, known = list(beta0 = 0),
       upperOpt <- c(upperOpt, noiseControl$g_bounds[2])
     }
     
+    
+    # environment storing values from log to pass to dlog
     out <- optim(par = parinit, fn = fn, gr = gr, method = "L-BFGS-B", lower = lowerOpt, upper = upperOpt, theta = known[["theta"]], 
-                 nu = known$nu, sigma2 = known$sigma2, g = known$g,
+                 nu = known$nu, sigma2 = known$sigma2, g = known$g, env = environment(),
                  X0 = X0, Z0 = Z0, Z = Z, mult = mult, beta0 = beta0, control = list(fnscale = -1, maxit = maxit))
     
     ## Post-processing
@@ -342,7 +357,7 @@ mleHomTP <- function(X, Z, lower, upper, known = list(beta0 = 0),
   
   psi_0 <- drop(crossprod(Z0 - beta0, Ki) %*% (Z0 - beta0))
   
-  psi <- drop(crossprod(Z - beta0, Z - beta0) - crossprod((Z0 - beta0) * mult, Z0 - beta0))/g_out + psi_0
+  psi <- drop(crossprod(Z - beta0) - crossprod((Z0 - beta0) * mult, Z0 - beta0))/g_out + psi_0
   
   res <- list(theta = theta_out, g = g_out, nu = nu_out, sigma2 = sigma2_out, mult = mult,
               # trendtype = trendtype,
@@ -423,7 +438,7 @@ predict.homTP <- function(object, x, xprime = NULL, ...){
   # if(object$trendtype == 'SK'){
   sd2 <- (object$nu + object$psi - 2) / (object$nu + n1 - 2) * as.vector(object$sigma2 - fast_diag(kx, tcrossprod(object$Ki, kx)))
   # }else{
-  # sd2 <- as.vector(object$nu2_hat - fast_diag(kx, tcrossprod(object$Ki, kx)) + (1 - tcrossprod(rowSums(object$Ki), kx))^2/sum(object$Ki))
+  # sd2 <- as.vector(object$nu_hat - fast_diag(kx, tcrossprod(object$Ki, kx)) + (1 - tcrossprod(rowSums(object$Ki), kx))^2/sum(object$Ki))
   # }
   
   ## In case of numerical errors, some sd2 values may become negative
@@ -437,9 +452,13 @@ predict.homTP <- function(object, x, xprime = NULL, ...){
   }
   
   if(!is.null(xprime)){
-    kxprime <- object$sigma2 * cov_gen(X1 = xprime, X2 = object$X, theta = object$theta, type = object$covtype)
+    kxprime <- object$sigma2 * cov_gen(X1 = object$X0, X2 = xprime, theta = object$theta, type = object$covtype)
     # if(object$trendtype == 'SK'){
-    cov <- (object$nu + object$psi - 2) / (object$nu + n1 - 2) *(object$sigma2 * cov_gen(X1 = x, X2 = xprime, theta = object$theta, type = object$covtype) - kx %*% tcrossprod(object$Ki, kxprime))
+    if(nrow(x) > nrow(xprime)){
+      cov <- (object$nu + object$psi - 2) / (object$nu + n1 - 2) * (object$sigma2 * cov_gen(X1 = x, X2 = xprime, theta = object$theta, type = object$covtype) - kx %*% (object$Ki %*% kxprime))
+    }else{
+      cov <- (object$nu + object$psi - 2) / (object$nu + n1 - 2) * (object$sigma2 * cov_gen(X1 = x, X2 = xprime, theta = object$theta, type = object$covtype) - (kx %*% object$Ki) %*% kxprime)
+    }
     # }else{
     # cov <- object$sigma2 * cov_gen(X1 = x, X2 = xprime, theta = object$theta, type = object$covtype) - kx %*% tcrossprod(object$Ki, kxprime) + crossprod(1 - tcrossprod(rowSums(object$Ki), kx), 1 - tcrossprod(rowSums(object$Ki), kxprime))/sum(object$Ki)
     # }
@@ -475,7 +494,7 @@ rebuild.homTP <- function(object, robust = FALSE){
 # dmt(Z, mean = rep(0,length(Z)), S = (model$nu - 2)/model$nu*(model$sigma2 * cov_gen(X, theta = model$theta, type = "Matern5_2") + diag(rep(model$Lambda, times = model$mult) + 1e-8)), df = model$nu,  log = T)
 logLik_HetTP <- function(X0, Z0, Z, mult, Delta, theta, g, nu, sigma2, k_theta_g = NULL, theta_g = NULL, logN = TRUE,
                          beta0 = 0, eps = sqrt(.Machine$double.eps), covtype = "Gaussian",
-                         penalty = T, hardpenalty = T){
+                         penalty = T, hom_ll = NULL, env = NULL, trace = 0){
   n <- nrow(X0)
   N <- length(Z)
   
@@ -502,10 +521,20 @@ logLik_HetTP <- function(X0, Z0, Z, mult, Delta, theta, g, nu, sigma2, k_theta_g
   
   LambdaN <- rep(Lambda, times = mult)
   
-  Ki <- chol(add_diag(sigma2 *  cov_gen(X1 = X0, theta = theta, type = covtype), Lambda/mult + eps))
+  C <- cov_gen(X1 = X0, theta = theta, type = covtype)
+  Ki <- chol(add_diag(sigma2 * C, Lambda/mult + eps))
   ldetKi <- - 2 * sum(log(diag(Ki))) # log determinant from Cholesky
   Ki <- chol2inv(Ki)
   
+  if(!is.null(env)){
+    env$C <- C
+    env$Cg <- Cg
+    env$Kg_c <- Kg_c
+    env$Kgi <- Kgi
+    env$ldetKi <- ldetKi
+    env$Ki <- Ki
+  }
+  
   psi_0 <- drop(crossprod(Z0 - beta0, Ki) %*% (Z0 - beta0))
   
   psi <- drop(crossprod((Z - beta0)/LambdaN, Z - beta0) - crossprod((Z0 - beta0) * mult/Lambda, Z0 - beta0) + psi_0)
@@ -513,22 +542,29 @@ logLik_HetTP <- function(X0, Z0, Z, mult, Delta, theta, g, nu, sigma2, k_theta_g
   loglik <- -N/2 * log((nu - 2) * pi) + ldetKi/2 - 1/2 * sum((mult - 1) * log(Lambda) + log(mult)) + lgamma((nu + N)/2) - lgamma(nu/2) - (nu + N)/2 * log(1 + psi/(nu - 2))
   
   if(penalty){
-    nu2_hat_var <- drop(crossprod(Delta - nmean, Kgi) %*% (Delta - nmean))/length(Delta)
+    nu_hat_var <- drop(crossprod(Delta - nmean, Kgi) %*% (Delta - nmean))/length(Delta)
     
     ## To avoid 0 variance, e.g., when Delta = nmean
-    if(nu2_hat_var < eps) return(loglik)
+    if(nu_hat_var < eps) return(loglik)
     
-    if(hardpenalty)
-      return(loglik + min(0, - n/2 * log(nu2_hat_var) - sum(log(diag(Kg_c))) - n/2*log(2*pi) - n/2))
+    # if(hardpenalty)
+    #   return(loglik + min(0, - n/2 * log(nu_hat_var) - sum(log(diag(Kg_c))) - n/2*log(2*pi) - n/2))
     
-    return(loglik - n/2 * log(nu2_hat_var) - sum(log(diag(Kg_c))) - n/2*log(2*pi) - n/2)
+    pen <- - n/2 * log(nu_hat_var) - sum(log(diag(Kg_c))) - n/2*log(2*pi) - n/2
+    
+    if(loglik < hom_ll && pen > 0){
+      if(trace> 0) warning("Penalty is desactivated when unpenalized likelihood is lower than its homTP equivalent")
+      return(loglik)
+    }
+    
+    return(loglik + pen)
   }
   return(loglik)
 }
 
 dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g = NULL, theta_g = NULL, beta0 = NULL, pX = NULL,
-                         logN = TRUE, SiNK = FALSE, components = NULL, eps = sqrt(.Machine$double.eps), covtype = "Gaussian", SiNK_eps = 1e-4,
-                         penalty = T, hardpenalty = T){
+                         logN = TRUE, components = NULL, eps = sqrt(.Machine$double.eps), covtype = "Gaussian", 
+                         penalty = T, hom_ll = NULL, env = NULL){
   
   ## Verifications
   if(is.null(k_theta_g) && is.null(theta_g))
@@ -554,9 +590,16 @@ dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g
   n <- nrow(X0)
   N <- length(Z)
   
-  Cg <- cov_gen(X1 = X0, theta = theta_g, type = covtype)
-  Kg_c <- chol(Cg + diag(eps + g/mult))
-  Kgi <- chol2inv(Kg_c)
+  if(!is.null(env)){
+    Cg <- env$Cg
+    Kg_c <- env$Kg_c
+    Kgi <- env$Kgi
+  }else{
+    Cg <- cov_gen(X1 = X0, theta = theta_g, type = covtype)
+    Kg_c <- chol(Cg + diag(eps + g/mult))
+    Kgi <- chol2inv(Kg_c)
+  }
+
   # M <- Cg %*% Kgi
   M <- add_diag(Kgi * (-eps - g / mult), rep(1, n))
   
@@ -570,11 +613,11 @@ dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g
   ## Precomputations for reuse
   KgiD <- Kgi %*% (Delta - nmean)
   
-  if(penalty){
-    nu2_hat_var <- drop(crossprod(KgiD, (Delta - nmean)))/length(Delta) 
-    # To prevent numerical issues when Delta = nmean, giving a positive penalty (or if the penalty is positive)
-    if(nu2_hat_var < eps || (hardpenalty && (- n/2 * log(nu2_hat_var) - sum(log(diag(Kg_c))) - n/2*log(2*pi) - n/2) > 0)) penalty <- FALSE
-  }                                          
+  # if(penalty){
+  #   nu_hat_var <- drop(crossprod(KgiD, (Delta - nmean)))/length(Delta) 
+  #   # To prevent numerical issues when Delta = nmean, giving a positive penalty (or if the penalty is positive)
+  #   if(nu_hat_var < eps || (hardpenalty && (- n/2 * log(nu_hat_var) - sum(log(diag(Kg_c))) - n/2*log(2*pi) - n/2) > 0)) penalty <- FALSE
+  # }                                          
   
   Lambda <- drop(nmean + M %*% (Delta - nmean))
   if(logN){
@@ -586,9 +629,17 @@ dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g
   
   LambdaN <- rep(Lambda, times = mult)
   
-  C <- cov_gen(X1 = X0, theta = theta, type = covtype)
-  Ki <- chol2inv(chol(sigma2 * C + diag(Lambda/mult + eps)))
-  
+  if(!is.null(env)){
+    C <- env$C
+    Ki <- env$Ki
+    ldetKi <- env$ldetKi
+  }else{
+    C <- cov_gen(X1 = X0, theta = theta, type = covtype)
+    Ki <- chol(sigma2 * C + diag(Lambda/mult + eps))
+    ldetKi <- - 2 * sum(log(diag(Ki))) # log determinant from Cholesky
+    Ki <- chol2inv(Ki)
+  }
+
   if(is.null(beta0))
     beta0 <- drop(colSums(Ki) %*% Z0 / sum(Ki))
   
@@ -600,6 +651,20 @@ dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g
   
   psi <- drop(crossprod((Z - beta0)/LambdaN, Z - beta0) - crossprod((Z0 - beta0) * mult/Lambda, Z0 - beta0) + psi_0)
   
+  if(penalty){
+    nu_hat_var <- drop(crossprod(KgiD, (Delta - nmean)))/length(Delta) 
+    
+    # To prevent numerical issues when Delta = nmean, resulting in divisions by zero
+    if(nu_hat_var < eps){
+      penalty <- FALSE
+    }else{
+      loglik <- -N/2 * log(2*pi) - N/2 * log(psi/N)  + 1/2 * ldetKi - 1/2 * sum((mult - 1) * log(Lambda) + log(mult)) - N/2
+      pen <- - n/2 * log(nu_hat_var) - sum(log(diag(Kg_c))) - n/2*log(2*pi) - n/2
+      if(hom_ll > loglik && pen > 0) penalty <- FALSE
+    }
+    # if(nu_hat_var < eps || (hardpenalty && (- n/2 * log(nu_hat_var) - sum(log(diag(Kg_c))) - n/2*log(2*pi) - n/2) > 0)) penalty <- FALSE
+  }   
+  
   dLogL_dtheta <- dLogL_dDelta <- dLogL_dkthetag <- dLogL_dthetag <- dLogL_dg <- dLogL_dnu <- dLogL_dsigma2 <- NULL
   
   # First component, derivative of logL with respect to theta
@@ -634,7 +699,7 @@ dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g
         dLogL_dtheta[i] <- dLogL_dtheta[i] - 1/2 * sum((mult - 1) * dLdtk/Lambda) # derivative of the sum(a_i - 1)log(lambda_i)
         
         if(penalty)
-          dLogL_dtheta[i] <- dLogL_dtheta[i]  + 1/2 * crossprod(KgiD, dCg_dthetak) %*% KgiD / nu2_hat_var  - 1/2 * fast_trace(Kgi, dCg_dthetak)
+          dLogL_dtheta[i] <- dLogL_dtheta[i]  + 1/2 * crossprod(KgiD, dCg_dthetak) %*% KgiD / nu_hat_var  - 1/2 * fast_trace(Kgi, dCg_dthetak)
         
       }else{
         dLogL_dtheta[i] <- sigma2 * (nu + N)/(2 * (nu + psi - 2)) * crossprod(KiZ0, dC_dthetak) %*% KiZ0  - sigma2/2 * fast_trace(Ki, dC_dthetak)
@@ -682,7 +747,7 @@ dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g
       dLogL_dthetag[i] <- crossprod(dCg_dthetagk %*% KgiD - M %*% (dCg_dthetagk %*% KgiD) -
                                       (1 - rsM) * drop(rSKgi %*% dCg_dthetagk %*% (Kgi %*% Delta) * sKgi - rSKgi %*% Delta * (rSKgi %*% dCg_dthetagk %*% rSKgi))/sKgi^2, dLogLdLambda) #chain rule
       # Penalty term
-      if(penalty) dLogL_dthetag[i] <- dLogL_dthetag[i] + 1/2 * crossprod(KgiD, dCg_dthetagk) %*% KgiD/nu2_hat_var - fast_trace(Kgi, dCg_dthetagk)/2 
+      if(penalty) dLogL_dthetag[i] <- dLogL_dthetag[i] + 1/2 * crossprod(KgiD, dCg_dthetagk) %*% KgiD/nu_hat_var - fast_trace(Kgi, dCg_dthetagk)/2 
     }
   }
   
@@ -701,13 +766,13 @@ dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g
   # Additional penalty terms on Delta
   if(penalty){
     if("Delta" %in% components){
-      dLogL_dDelta <- dLogL_dDelta - KgiD / nu2_hat_var
+      dLogL_dDelta <- dLogL_dDelta - KgiD / nu_hat_var
     }
     if("k_theta_g" %in% components){
-      dLogL_dkthetag <- dLogL_dkthetag + 1/2 * crossprod(KgiD, dCg_dk) %*% KgiD / nu2_hat_var - fast_trace(Kgi, dCg_dk)/2 
+      dLogL_dkthetag <- dLogL_dkthetag + 1/2 * crossprod(KgiD, dCg_dk) %*% KgiD / nu_hat_var - fast_trace(Kgi, dCg_dk)/2 
     }
     if("g" %in% components){
-      dLogL_dg <- dLogL_dg + 1/2 * crossprod(KgiD/mult, KgiD) / nu2_hat_var - sum(diag(Kgi)/mult)/2 
+      dLogL_dg <- dLogL_dg + 1/2 * crossprod(KgiD/mult, KgiD) / nu_hat_var - sum(diag(Kgi)/mult)/2 
     }
     
   }
@@ -757,7 +822,7 @@ dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g
 ##'   \item \code{logN}, when \code{TRUE} (default), the log-noise process is modeled.
 ##'   \item \code{initStrategy} one of \code{'simple'}, \code{'residuals'} (default) and \code{'smoothed'} to obtain starting values for \code{Delta}, see Details
 ##'   \item \code{penalty} when \code{TRUE}, the penalized version of the likelihood is used (i.e., the sum of the log-likelihoods of the mean and variance processes, see References).
-##'   \item \code{hardpenalty} is \code{TRUE}, the log-likelihood from the noise GP is taken into account only if negative.
+## '   \item \code{hardpenalty} is \code{TRUE}, the log-likelihood from the noise GP is taken into account only if negative.
 ##'   \item \code{checkHom} when \code{TRUE}, if the log-likelihood with a homoskedastic model is better, then return it.
 ##'   \item \code{trace} optional scalar (default to \code{0}). If positive, tracing information on the fitting process.
 ##' If \code{1}, information is given about the result of the heterogeneous model optimization.
@@ -771,7 +836,7 @@ dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g
 ##' \itemize{
 ##' \item \code{theta} lengthscale parameter(s) for the mean process either one value (isotropic) or a vector (anistropic)
 ##' \item \code{Delta} vector of nuggets corresponding to each design in \code{X0}, that are smoothed to give \code{Lambda}
-##' (as the global covariance matrix depend on \code{Delta} and \code{nu2_hat}, it is recommended to also pass values for \code{theta})
+##' (as the global covariance matrix depend on \code{Delta} and \code{nu_hat}, it is recommended to also pass values for \code{theta})
 ##' \item \code{beta0} constant trend of the mean process
 ##' \item \code{k_theta_g} constant used for link mean and noise processes lengthscales, when \code{settings$linkThetas == 'joint'}
 ##' \item \code{theta_g} either one value (isotropic) or a vector (anistropic) for lengthscale parameter(s) of the noise process, when \code{settings$linkThetas != 'joint'}
@@ -808,7 +873,7 @@ dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g
 ##' while \code{init$g_H} may be use to pass an initial nugget value.
 ##' The resulting lengthscales provide initial values for \code{theta} (or update them if given in \code{init}). \cr \cr
 ##' If necessary, a second homoskedastic model, \code{modNugs}, is fitted to the empirical residual variance between the prediction
-##'  given by \code{modHom} at \code{X0} and \code{Z} (up to \code{modHom$nu2_hat}).
+##'  given by \code{modHom} at \code{X0} and \code{Z} (up to \code{modHom$nu_hat}).
 ##' Note that when specifying \code{settings$linkThetas == 'joint'}, then this second homoskedastic model has fixed lengthscale parameters.
 ##' Starting values for \code{theta_g} and \code{g} are extracted from \code{modNugs}.\cr \cr
 ##' Finally, three initialization schemes for \code{Delta} are available with \code{settings$initStrategy}: 
@@ -826,7 +891,7 @@ dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g
 ##' \item \code{theta}: unless given, maximum likelihood estimate (mle) of the lengthscale parameter(s),
 ##' \item \code{Delta}: unless given, mle of the nugget vector (non-smoothed),
 ##' \item \code{Lambda}: predicted input noise variance at \code{X0}, 
-##' \item \code{nu2_hat}: plugin estimator of the variance,
+##' \item \code{nu_hat}: plugin estimator of the variance,
 ##' \item \code{theta_g}: unless given, mle of the lengthscale(s) of the noise/log-noise process,
 ##' \item \code{k_theta_g}: if \code{settings$linkThetas == 'joint'}, mle for the constant by which lengthscale parameters of \code{theta} are multiplied to get \code{theta_g},
 ##' \item \code{g}: unless given, mle of the nugget of the noise/log-noise process,
@@ -838,7 +903,7 @@ dlogLikHetTP <- function(X0, Z0, Z, mult, Delta, theta, nu, sigma2, g, k_theta_g
 ##' \item \code{nit_opt}, \code{msg}: \code{counts} and \code{message} returned by \code{\link[stats]{optim}}
 ##' \item \code{modHom}: homoskedastic GP model of class \code{homGP} used for initialization of the mean process,
 ##' \item \code{modNugs}: homoskedastic GP model of class \code{homGP} used for initialization of the noise/log-noise process,
-##' \item \code{nu2_hat_var}: variance of the noise process,
+##' \item \code{nu_hat_var}: variance of the noise process,
 ##' \item \code{used_args}: list with arguments provided in the call to the function, which is saved in \code{call},
 ##' \item \code{X0}, \code{Z0}, \code{Z}, \code{eps}, \code{logN}, \code{covtype}: values given in input
 ##'}
@@ -1027,9 +1092,9 @@ mleHetTP <- function(X, Z, lower, upper,
   if(!is.null(settings$penalty))
     penalty <- settings$penalty
   
-  hardpenalty <- FALSE
-  if(!is.null(settings$hardpenalty))
-    hardpenalty <- settings$hardpenalty
+  # hardpenalty <- FALSE
+  # if(!is.null(settings$hardpenalty))
+  #   hardpenalty <- settings$hardpenalty
   
   if(is.null(settings$checkHom))
     settings$checkHom <- TRUE
@@ -1104,6 +1169,10 @@ mleHetTP <- function(X, Z, lower, upper,
   if(is.null(known$beta0))  known$beta0 <- 0
   beta0 <- known$beta0
   
+  if(is.null(components) && is.null(known$theta_g)){
+    known$theta_g <- known$k_theta_g * known$theta
+  }
+  
   if(is.null(noiseControl$g_bounds)){
     noiseControl$g_bounds <- c(1e-6, 0.1)
   }
@@ -1116,18 +1185,18 @@ mleHetTP <- function(X, Z, lower, upper,
     noiseControl$sigma2_bounds <- c(sqrt(.Machine$double.eps), 1e4)
   }
   
-  ## Advanced option Single Nugget Kriging model for the noise process
-  if(!is.null(noiseControl$SiNK) && noiseControl$SiNK){
-    SiNK <- TRUE
-    if(is.null(noiseControl$SiNK_eps)){
-      SiNK_eps <- 1e-4
-    }else{
-      SiNK_eps <- noiseControl$SiNK_eps
-    }
-  }else{
-    SiNK <- FALSE
-    SiNK_eps <- 1e-4
-  } 
+  # ## Advanced option Single Nugget Kriging model for the noise process
+  # if(!is.null(noiseControl$SiNK) && noiseControl$SiNK){
+  #   SiNK <- TRUE
+  #   if(is.null(noiseControl$SiNK_eps)){
+  #     SiNK_eps <- 1e-4
+  #   }else{
+  #     SiNK_eps <- noiseControl$SiNK_eps
+  #   }
+  # }else{
+  #   SiNK <- FALSE
+  #   SiNK_eps <- 1e-4
+  # } 
   
   
   ### Automatic Initialisation
@@ -1173,7 +1242,7 @@ mleHetTP <- function(X, Z, lower, upper,
       rKI <- FALSE
     } 
     modHom <- mleHomTP(X = list(X0 = X0, Z0 = Z0, mult = mult), Z = Z, lower = lower,
-                       known = list(theta = known$theta, g = known$g_H, nu = known$nu, sigma2 = known$sigma2),
+                       known = list(theta = known$theta, g = known$g_H, nu = known$nu, sigma2 = known$sigma2, beta0 = beta0),
                        init = list(theta = init$theta, g = g_init, nu = init$nu, sigma2 = init$sigma2),
                        upper = upper, covtype = covtype, maxit = maxit,
                        noiseControl = list(g_bounds = c(noiseControl$g_min, noiseControl$g_max),
@@ -1184,13 +1253,12 @@ mleHetTP <- function(X, Z, lower, upper,
     
     if(is.null(known$sigma2)) init$sigma2 <- modHom$sigma2
     
-    if(is.null(known$nu)) init$nu <- modHom$nu
+    if(is.null(known$nu) && is.null(init$nu)) init$nu <- modHom$nu # Estimating nu may be hard, and even more for homTP
     
     if(is.null(init$Delta)){
-      predHom <- predict(x = X0, object = modHom)$mean
+      predHom <- suppressWarnings(predict(x = X0, object = modHom)$mean)
       nugs_est <- (Z - rep(predHom, times = mult))^2 #squared deviation from the homoskedastic prediction mean to the actual observations
-      nugs_est <-  nugs_est  # to be homegeneous with Delta
-      
+
       if(logN){
         nugs_est <- log(nugs_est)
       }
@@ -1205,6 +1273,16 @@ mleHetTP <- function(X, Z, lower, upper,
       noiseControl$lowerTheta_g <- modHom$theta
     }
     
+    if(settings$initStrategy == 'simple'){
+      if(logN){
+        init$Delta <- rep(log(modHom$g), nrow(X0))
+      }else{
+        init$Delta <- rep(modHom$g, nrow(X0))
+      }
+    }
+    if(settings$initStrategy == 'residuals')
+      init$Delta <- nugs_est0
+    
   }
   
   
@@ -1252,9 +1330,10 @@ mleHetTP <- function(X, Z, lower, upper,
                           lower = noiseControl$lowerTheta_g, upper = noiseControl$upperTheta_g,
                           init = list(theta = init$theta_g, g =  init$g), covtype = covtype, noiseControl = noiseControl,
                           maxit = maxit, eps = eps, settings = list(return.Ki = F))
-      prednugs <- predict(x = X0, object = modNugs)
+      prednugs <- suppressWarnings(predict(x = X0, object = modNugs))
       
     }else{
+      if(!exists("nugs_est0")) nugs_est0 <- init$Delta
       
       if(is.null(init$g)){
         init$g <- 0.05
@@ -1265,7 +1344,7 @@ mleHetTP <- function(X, Z, lower, upper,
                           lower = noiseControl$lowerTheta_g, upper = noiseControl$upperTheta_g,
                           init = list(theta = init$theta_g, g =  init$g), covtype = covtype, noiseControl = noiseControl,
                           maxit = maxit, eps = eps, settings = list(return.Ki = F))
-      prednugs <- predict(x = X0, object = modNugs)
+      prednugs <- suppressWarnings(predict(x = X0, object = modNugs))
     }
     
     if(settings$initStrategy == 'simple'){
@@ -1301,7 +1380,7 @@ mleHetTP <- function(X, Z, lower, upper,
   
   ### Start of optimization of the log-likelihood
   fn <- function(par, X0, Z0, Z, mult, Delta = NULL, theta = NULL, g = NULL, k_theta_g = NULL, theta_g = NULL,
-                 nu = NULL, sigma2 = NULL, logN = FALSE, beta0 = NULL){
+                 nu = NULL, sigma2 = NULL, logN = FALSE, beta0 = NULL, hom_ll, env = NULL){
     
     idx <- 1 # to store the first non used element of par
     
@@ -1339,12 +1418,12 @@ mleHetTP <- function(X, Z, lower, upper,
     
     return(logLik_HetTP(X0 = X0, Z0 = Z0, Z = Z, mult = mult, Delta = Delta, theta = theta, g = g, k_theta_g = k_theta_g, theta_g = theta_g,
                         nu = nu, sigma2 = sigma2, logN = logN, beta0 = beta0, covtype = covtype, eps = eps,
-                        penalty = penalty, hardpenalty = hardpenalty))
+                        penalty = penalty, hom_ll = hom_ll, env = env, trace = trace))
   }
   
   
   gr <- function(par, X0, Z0, Z, mult, Delta = NULL, theta = NULL, g = NULL, k_theta_g = NULL, theta_g = NULL, nu = NULL, sigma2 = NULL, logN = FALSE,
-                 beta0 = NULL, pX = NULL){
+                 beta0 = NULL, pX = NULL, hom_ll, env = NULL){
     
     idx <- 1 # to store the first non used element of par
     
@@ -1388,8 +1467,8 @@ mleHetTP <- function(X, Z, lower, upper,
     # grad(fn, x = par, X0 = X0, Z0 = Z0, Z = Z, mult = mult, logN = logN, beta0 = beta0, method.args=list(r = 6))
     
     return(dlogLikHetTP(X0 = X0, Z0 = Z0, Z = Z, mult = mult, Delta = Delta, theta = theta, g = g, k_theta_g = k_theta_g, theta_g = theta_g,
-                        nu = nu, sigma2 = sigma2, logN = logN, beta0 = beta0, components = components, covtype = covtype, eps = eps, SiNK_eps = SiNK_eps,
-                        penalty = penalty, hardpenalty = hardpenalty))
+                        nu = nu, sigma2 = sigma2, logN = logN, beta0 = beta0, components = components, covtype = covtype, eps = eps,
+                        penalty = penalty, hom_ll = hom_ll, env = env))
     
   }
   
@@ -1487,9 +1566,22 @@ mleHetTP <- function(X, Z, lower, upper,
   mle_par <- known # Store infered and known parameters
   if(!is.null(components)){
     
+    if(!is.null(modHom)){
+      hom_ll <- modHom$ll
+    }else{
+      ## Compute reference homoskedastic likelihood, with fixed theta for speed
+      modHom_tmp <- mleHomTP(X = list(X0 = X0, Z0 = Z0, mult = mult), Z = Z, lower = lower, upper = upper, upper,
+                             known = list(theta = known[["theta"]], g = known$g_H, nu = known$nu, sigma2 = known$sigma2, beta0 = 0),
+                             covtype = covtype, init = init,
+                             noiseControl = noiseControl, eps = eps)
+      
+      hom_ll <- modHom_tmp$ll
+    } 
+    
     ## Maximization of the log-likelihood
-    out <- optim(par = parinit, fn = fn, gr = gr, method = "L-BFGS-B", lower = lowerOpt, upper = upperOpt, X0 = X0, Z0 = Z0, Z = Z,
-                 mult = mult, logN = logN, Delta = known$Delta, theta = known$theta, g = known$g, k_theta_g = known$k_theta_g, theta_g = known$theta_g,
+    out <- optim(par = parinit, fn = fn, gr = gr, method = "L-BFGS-B", lower = lowerOpt, upper = upperOpt,
+                 X0 = X0, Z0 = Z0, Z = Z, mult = mult, logN = logN, Delta = known$Delta, theta = known$theta,
+                 g = known$g, k_theta_g = known$k_theta_g, theta_g = known$theta_g, hom_ll = hom_ll, env = environment(),
                  nu = known$nu, beta0 = known$beta0, sigma2 = known$sigma2, control = list(fnscale = -1, maxit = maxit))
     
     ## Temporary
@@ -1503,11 +1595,12 @@ mleHetTP <- function(X, Z, lower, upper,
     
     ## Post-processing
     idx <- 1
-    if(is.null(known$theta)){
+    if(is.null(known[["theta"]])){
       mle_par$theta <- out$par[1:length(init$theta)]
       idx <- idx + length(init$theta)
       if(trace > 1) cat("Theta |", mle_par$theta, " | ", lower, " | ", upper, "\n")
     }
+
     if(is.null(known$Delta)){
       mle_par$Delta <- out$par[idx:(idx - 1 + length(init$Delta))]
       idx <- idx + length(init$Delta)
@@ -1518,8 +1611,8 @@ mleHetTP <- function(X, Z, lower, upper,
           if(!logN) cat("Delta |", mle_par$Delta[i_tmp], " | ", pmax(mult[i_tmp] * eps, init$Delta[i_tmp] / 1000), " | ", init$Delta[i_tmp] * 100, "\n")
         }
       }
-      
     }
+    
     if(jointThetas){
       if(is.null(known$k_theta_g)){
         mle_par$k_theta_g <- out$par[idx]
@@ -1565,7 +1658,7 @@ mleHetTP <- function(X, Z, lower, upper,
   if(penalty){
     ll_non_pen <- logLik_HetTP(X0 = X0, Z0 = Z0, Z = Z, mult = mult, Delta = mle_par$Delta, theta = mle_par$theta, g = mle_par$g,
                                k_theta_g = mle_par$k_theta_g, theta_g = mle_par$theta_g, sigma2 = mle_par$sigma2, nu = mle_par$nu,
-                               logN = logN, beta0 = mle_par$beta0, covtype = covtype, eps = eps, penalty = FALSE)
+                               logN = logN, beta0 = mle_par$beta0, covtype = covtype, eps = eps, penalty = FALSE, trace = trace)
   }else{
     ll_non_pen <- out$value
   }
@@ -1589,14 +1682,14 @@ mleHetTP <- function(X, Z, lower, upper,
   
   nmean <- drop(rowSums(Kgi) %*% mle_par$Delta / sum(Kgi)) ## ordinary kriging mean
   
-  nu2_hat_var <- max(eps, drop(crossprod(mle_par$Delta - nmean, Kgi) %*% (mle_par$Delta - nmean))/length(mle_par$Delta))
+  nu_hat_var <- max(eps, drop(crossprod(mle_par$Delta - nmean, Kgi) %*% (mle_par$Delta - nmean))/length(mle_par$Delta))
   
-  if(SiNK){
-    rhox <- 1 / rho_AN(xx = X0, X0 = mle_par$pX, theta_g = mle_par$theta_g, g = mle_par$g, type = covtype, eps = eps, SiNK_eps = SiNK_eps, mult = mult)
-    M <-  rhox * Cg %*% (Kgi %*% (mle_par$Delta - nmean))
-  }else{
-    M <- Cg %*% (Kgi %*% (mle_par$Delta - nmean))
-  }
+  # if(SiNK){
+  #   rhox <- 1 / rho_AN(xx = X0, X0 = mle_par$pX, theta_g = mle_par$theta_g, g = mle_par$g, type = covtype, eps = eps, SiNK_eps = SiNK_eps, mult = mult)
+  #   M <-  rhox * Cg %*% (Kgi %*% (mle_par$Delta - nmean))
+  # }else{
+  M <- Cg %*% (Kgi %*% (mle_par$Delta - nmean))
+  # }
   # }else{
   #   Cg <- cov_gen(X1 = mle_par$pX, theta = mle_par$theta_g, type = covtype)
   #   Kgi <- chol2inv(chol(Cg + diag(eps + mle_par$g/mult)))
@@ -1605,7 +1698,7 @@ mleHetTP <- function(X, Z, lower, upper,
   #   
   #   nmean <- drop(rowSums(Kgi) %*% mle_par$Delta / sum(Kgi)) ## ordinary kriging mean
   #   
-  #   nu2_hat_var <- max(eps, drop(crossprod(mle_par$Delta - nmean, Kgi) %*% (mle_par$Delta - nmean))/length(mle_par$Delta))
+  #   nu_hat_var <- max(eps, drop(crossprod(mle_par$Delta - nmean, Kgi) %*% (mle_par$Delta - nmean))/length(mle_par$Delta))
   #   
   #   if(SiNK){
   #     rhox <- 1 / rho_AN(xx = X0, X0 = mle_par$pX, theta_g = mle_par$theta_g, g = mle_par$g, type = covtype, eps = eps, SiNK_eps = SiNK_eps, mult = mult)
@@ -1640,9 +1733,9 @@ mleHetTP <- function(X, Z, lower, upper,
   
   res <- list(theta = mle_par$theta, Delta = mle_par$Delta, psi = as.numeric(psi), beta0 = mle_par$beta0,
               k_theta_g = mle_par$k_theta_g, theta_g = mle_par$theta_g, g = mle_par$g, nmean = nmean, Lambda = Lambda,
-              ll = out$value, ll_non_pen = ll_non_pen, nit_opt = out$counts, logN = logN, SiNK = SiNK, covtype = covtype, pX = mle_par$pX, msg = out$message,
-              X0 = X0, Z0 = Z0, Z = Z, mult = mult, trendtype = trendtype, SiNK_eps = SiNK_eps, eps = eps,
-              nu2_hat_var = nu2_hat_var, call = match.call(), nu = mle_par$nu, sigma2 = mle_par$sigma2,
+              ll = out$value, ll_non_pen = ll_non_pen, nit_opt = out$counts, logN = logN, covtype = covtype, pX = mle_par$pX, msg = out$message,
+              X0 = X0, Z0 = Z0, Z = Z, mult = mult, trendtype = trendtype, eps = eps,
+              nu_hat_var = nu_hat_var, call = match.call(), nu = mle_par$nu, sigma2 = mle_par$sigma2,
               used_args = list(noiseControl = noiseControl, settings = settings, lower = lower, upper = upper, known = known))
   
   if(settings$return.matrices){
@@ -1715,9 +1808,9 @@ predict.hetTP <- function(object, x, noise.var = FALSE, xprime = NULL, nugs.only
   }
   
   # if(noise.var){
-  #   if(is.null(object$nu2_hat_var))
-  #     object$nu2_hat_var <- max(object$eps, drop(crossprod(object$Delta - object$nmean, object$Kgi) %*% (object$Delta - object$nmean))/length(object$Delta)) ## To avoid 0 variance
-  #   sd2var = object$nu2_hat * object$nu2_hat_var* drop(1 - fast_diag(kg, tcrossprod(object$Kgi, kg)) + (1 - tcrossprod(rowSums(object$Kgi), kg))^2/sum(object$Kgi))
+  #   if(is.null(object$nu_hat_var))
+  #     object$nu_hat_var <- max(object$eps, drop(crossprod(object$Delta - object$nmean, object$Kgi) %*% (object$Delta - object$nmean))/length(object$Delta)) ## To avoid 0 variance
+  #   sd2var = object$nu_hat * object$nu_hat_var* drop(1 - fast_diag(kg, tcrossprod(object$Kgi, kg)) + (1 - tcrossprod(rowSums(object$Kgi), kg))^2/sum(object$Kgi))
   # }else{
   #   sd2var = NULL
   # }
@@ -1741,9 +1834,14 @@ predict.hetTP <- function(object, x, noise.var = FALSE, xprime = NULL, nugs.only
   
   
   if(!is.null(xprime)){
-    kxprime <- object$sigma2 * cov_gen(X1 = xprime, X2 = object$X0, theta = object$theta, type = object$covtype)
+    kxprime <- object$sigma2 * cov_gen(X1 = object$X0, X2 = xprime, theta = object$theta, type = object$covtype)
     # if(object$trendtype == 'SK'){
-    cov <- (object$nu + object$psi - 2) / (object$nu + n1 - 2) * (object$sigma2 * cov_gen(X1 = x, X2 = xprime, theta = object$theta, type = object$covtype) - kx %*% tcrossprod(object$Ki, kxprime))
+    if(nrow(x) > nrow(xprime)){
+      cov <- (object$nu + object$psi - 2) / (object$nu + n1 - 2) * (object$sigma2 * cov_gen(X1 = x, X2 = xprime, theta = object$theta, type = object$covtype) - kx %*% object$Ki %*% kxprime)
+    }else{
+      cov <- (object$nu + object$psi - 2) / (object$nu + n1 - 2) * (object$sigma2 * cov_gen(X1 = x, X2 = xprime, theta = object$theta, type = object$covtype) - (kx %*% object$Ki) %*% kxprime)
+    }
+    
     # }else{
     #   cov <- (object$nu + object$psi - 2) / (object$nu + n1 - 2) * (object$sigma2 * cov_gen(X1 = x, X2 = xprime, theta = object$theta, type = object$covtype) - kx %*% tcrossprod(object$Ki, kxprime) + crossprod(1 - tcrossprod(rowSums(object$Ki), kx), 1 - tcrossprod(rowSums(object$Ki), kxprime))/sum(object$Ki))
     # }
